<?php

/**
 * @package PandaRose_UI
 * @version 1.0.0
 */
/*
Plugin Name: PandaRose UI
Plugin URI: http://wordpress.org/plugins/pandarose-ui/
Description: An embeddable rendition of the PandaRose UI Foundation System
Author: Stefen Enns
Version: 1.0.0
*/

if (defined('PANDA_UI_VERSION'))
    return;

require_once(__DIR__ . '/vendor/autoload.php');

define('PANDA_UI_VERSION', '1.0.11');

use buzzingpixel\twigswitch\SwitchTwigExtension;
use Symfony\Component\Yaml\Yaml;
use ScssPhp\ScssPhp\Compiler;

class PandaRoseUIProxy
{
    protected $data = [];
    public $last = [];

    public function __get($key)
    {
        return isset($this->data[$key]) ? $this->data[$key] : null;
    }
    public function __set($key, $val)
    {
        $this->data[$key] = $val;
    }
    public function __isset($key)
    {
        return isset($this->data[$key]);
    }
}

class PandaRoseData implements ArrayAccess
{
    protected $data = [];

    public function __construct($data = [])
    {
        foreach ($data as $k => $v) {
            if (is_numeric($k)) {
                $this->data[$v] = true;
            } else {
                $this->data[$k] = $v;
            }
        }
    }

    public function offsetExists(mixed $offset): bool
    {
        return isset($this->data[$offset]);
    }

    public function offsetGet(mixed $offset): mixed
    {
        return isset($this->data[$offset]) ? $this->data[$offset] : null;
    }

    public function offsetSet(mixed $offset, mixed $value): void
    {
    }

    public function offsetUnset(mixed $offset): void
    {
    }

    public function __get($key)
    {
        return isset($this->data[$key]) ? $this->data[$key] : null;
    }

    public function __isset($key)
    {
        return isset($this->data[$key]);
    }

    public function __invoke(...$arguments)
    {
        $out = '';

        foreach ($arguments as $arg) {
            if (isset($this->data[$arg])) {
                $val = $this->data[$arg];
                if ($val === true) $val = 'true';
                if ($val === false) $val = 'false';
                if ($val === null) continue;
                $out .= ' data-' . $arg . '="' . addslashes($val) . '"';
            }
        }

        return trim($out);
    }
}

class PandaRoseUI
{
    private static $enqueued = false;
    public static $namespace = [];
    public static $deep = 0;
    public static $last = null;
    public static $in = [];
    public static $i = 0;
    public static $test;
    public static $twig;
    public static $tmpl = [];
    public static $repl = [];
    public static $that = [];
    public static $final = [];

    public static function getThis($filepath)
    {
        if (!isset(self::$that[$filepath]))
            self::$that[$filepath] = new PandaRoseUIProxy;
        return self::$that[$filepath];
    }

    public static function enqueue()
    {
        if (!static::$enqueued) {
            static::$enqueued = true;
            // wp_enqueue_script('main', 'https://cdn.jsdelivr.net/gh/pandarose/ui-dist@' . PANDA_UI_VERSION . '/index.bundle.js');
            // wp_enqueue_style('main', 'https://cdn.jsdelivr.net/gh/pandarose/ui-dist@' . PANDA_UI_VERSION . '/index.bundle.css');
            wp_enqueue_script('main', 'http://localhost/build/index.bundle.js');
            wp_enqueue_style('main', 'http://localhost/build/index.bundle.css');
        }
    }

    public static function after($callable)
    {
        add_filter(
            'final_output',
            $callable,
            PHP_INT_MAX,
            2
        );
    }

    public static function closeBuffer()
    {
        $out = '';

        $levels = ob_get_level();

        for ($i = 0; $i < $levels; $i++) {
            $out .= ob_get_clean();
        }

        return $out;
    }

    public static function sendBuffer()
    {
        $out = static::closeBuffer();

        $r = count(PandaRoseUI::$repl);

        for ($i = 0; $i < $r; $i++) {
            foreach (PandaRoseUI::$repl as $key => $value) {
                if (substr_count($out, $key)) {
                    $out = do_shortcode($out);
                    $out = str_replace($key, $value, $out);
                    unset(PandaRoseUI::$repl[$key]);
                }
            }
        }

        $scripts = '';

        foreach (PandaRoseUI::$final as $script) {
            $scripts .= '<script>' .
                'var jq = jQuery;' .
                'jq(document).ready(function() {' . $script . '});' .
                '</script>';
        }

        if (substr_count($out, '</body>')) {
            $out = str_replace('</body>', $scripts . '</body>', $out);
        } else {
            $out .= $scripts;
        }

        echo apply_filters('final_output', $out);
    }

    public static function keyString($key, $value, &$out)
    {
        $keys = explode('.', $key);
        $row = &$out;

        for ($i = 0, $end = count($keys) - 1; $i <= $end; $i++) {
            $cur = $keys[$i];
            if (!isset($row[$cur])) {
                if (isset($keys[$i + 1])) {
                    $row[$cur] = [];
                    $row = &$row[$cur];
                } else {
                    $row = &$row[$cur];
                }
            } else {
                $row = &$row[$cur];
            }
        }

        $row = $value;
    }
}

class PandaTwigLoader implements Twig_LoaderInterface
{
    protected $content;

    public function __construct($content)
    {
        $this->content = $content;
    }
    public function getSourceContext($name)
    {
        if ($name[0] === '#') {
            return new Twig_Source(PandaRoseUI::$tmpl[$name], $name, $name);
        }

        $content = '';

        $content = pandarose($name);

        $repl = md5(random_bytes(32));

        PandaRoseUI::$repl[$repl] = $content;

        return new Twig_Source($repl, $name, $name);
    }

    public function getCacheKey($name)
    {
        return $name;
    }

    public function isFresh($name, $time)
    {
        return true;
    }

    public function exists($name)
    {
        if ($name === '__loaded') {
            return true;
        }

        return true;
    }
}

function parseTwig($content, $data = [])
{
    if (!PandaRoseUI::$twig) {
        $dump = new Twig_Function(
            'dump',
            function ($var) {
                var_dump($var);
                return '';
            }
        );
        $slug = new Twig_Function(
            'slug',
            function ($str) {
                return sanitize_title($str);
            }
        );
        $post = new Twig_Function(
            'post',
            function ($key) {
                return isset(PandaRoseUI::$in[$key])
                    ? PandaRoseUI::$in[$key] : null;
            }
        );
        $load = new PandaTwigLoader($content);
        PandaRoseUI::$twig = $twig = new Twig_Environment($load, []);
        $twig->addFunction($dump);
        $twig->addFunction($slug);
        $twig->addFunction($post);
        $twig->addExtension(new SwitchTwigExtension());
    } else {
        $twig = PandaRoseUI::$twig;
    }

    PandaRoseUI::$i++;
    PandaRoseUI::$tmpl['#' . PandaRoseUI::$i] = $content;

    $html = $twig->render('#' . PandaRoseUI::$i, $data);

    return $html;
}

function pandarose($lookup = 'default', $data = null, $inner = '')
{
    if ($data === null && is_array($lookup)) {
        $data = $lookup;
        $lookup = isset($data['template']) ? $data['template'] : null;
    }

    if ($data === null)
        $data = [];

    if (PandaRoseUI::$last === $lookup)
        return '';

    if (isset($lookup) and $lookup) {
        $dir = null;
        if (substr_count($lookup, '/')) {
            list($dir, $template) = explode('/', $lookup, 2);
        } else {
            $template = $lookup;
        }
        if ($dir !== 'core') {
            $scan = [];

            if (!$dir) {
                foreach (PandaRoseUI::$namespace as $ns) {
                    $scan[] = ['pandarose-' . $ns, $template];
                    if (!substr_count($template, '.')) {
                        $scan[] = ['pandarose-' . $ns, $template . '.yml'];
                        $scan[] = ['pandarose-' . $ns, $template . '.md'];
                        $scan[] = ['pandarose-' . $ns, $template . '.html'];
                    }
                }
            }
            if (!$dir) {
                $scan[] = ['pandarose', $template];
                if (!substr_count($template, '.')) {
                    $scan[] = ['pandarose', $template . '.yml'];
                    $scan[] = ['pandarose', $template . '.md'];
                    $scan[] = ['pandarose', $template . '.html'];
                }
            } else {
                $scan[] = ['pandarose-' . $dir, $template];
                if (!substr_count($template, '.')) {
                    $scan[] = ['pandarose-' . $dir, $template . '.yml'];
                    $scan[] = ['pandarose-' . $dir, $template . '.md'];
                    $scan[] = ['pandarose-' . $dir, $template . '.html'];
                }
            }
            foreach ($scan as $row) {
                $post = get_page_by_title($row[1], 'OBJECT', $row[0]);
                if ($post) {
                    PandaRoseUI::$namespace[] = $dir;
                    PandaRoseUI::$last = $lookup;
                    $out = pandarose_template(null, $post->post_content, $data, $post->post_title, $inner);
                    array_pop(PandaRoseUI::$namespace);
                    PandaRoseUI::$last = null;
                    return $out;
                }
            }
        } else {
            $lookup = $template;
        }
        if ($template = trim(strtolower($lookup))) {
            $plugin = null;
            if (substr_count($template, '/')) {
                list($plugin, $template) = explode('/', $template);
            }
            if ($template = (preg_replace('/[^_a-z\d]/i', '', $template))) {
                $plugin = '';
                if ($plugin = (preg_replace('/[^_a-z\d]/i', '', $plugin))) {
                    $filepath = __DIR__ . DIRECTORY_SEPARATOR . '..' . DIRECTORY_SEPARATOR . $plugin . DIRECTORY_SEPARATOR . 'pandarose' . DIRECTORY_SEPARATOR . $template . '.php';
                } else {
                    $filepath = __DIR__ . DIRECTORY_SEPARATOR . 'templates' . DIRECTORY_SEPARATOR . $template . '.php';
                }
            }
            if (file_exists($filepath)) {
                PandaRoseUI::$namespace[] = $plugin;
                PandaRoseUI::$last = $lookup;
                $out = pandarose_template($filepath, null, $data, null, $inner);
                array_pop(PandaRoseUI::$namespace);
                PandaRoseUI::$last = null;
                return $out;
            }
        }
    }

    return '';
}

function pandarose_template(
    $filepath,
    $content = null,
    $data = [],
    $name = null,
    $inner = '',
) {
    PandaRoseUI::$deep++;

    if ($content || ($filepath and file_exists($filepath))) {
        $newHtml = null;
        $in = [];
        if ($json = ($_SERVER['HTTP_ACCEPT'] === 'application/json')) {
            if ($body = file_get_contents('php://input'))
                if (!is_array($in = json_decode($body, true)))
                    $in = [];
            PandaRoseUI::$in = $in;
        }
        if ($filepath) {
            ob_start();
            $that = PandaRoseUI::getThis($filepath);
            $d = $data;
            if (isset($d['d'])) unset($d['d']);
            $run = function () use ($d, $inner, $filepath) {
                PandaRoseUI::$i++;
                $id = 'e' . md5($filepath) . PandaRoseUI::$i;

                if (is_array($d)) {
                    if (!isset($d['id'])) $d['id'] = $id;
                    extract($d);
                }

                $data = new PandaRoseData($d);

                $result = [];

                if (!$res = require($filepath)) {
                    if ($res !== false)
                        $res = [];
                }

                if ($res === false) return false;

                return is_array($res) ? array_merge($result, $res) : $res;
            };

            $run = $run->bindTo($that, $that);
            $result = $run();

            $output = ob_get_clean();

            if ($result === false) {
                $output = '';
                $result = [];
            }
        } else {
            $result = [];
            if ($name) {
                $extn = pathinfo($name, PATHINFO_EXTENSION);
                $json = null;
                switch ($extn) {
                    case 'md':
                        $lines = explode("\n", $content);
                        $content = '';
                        foreach ($lines as $line) {
                            if ('include: ' === substr($line, 0, 9)) {
                                $incl = trim(substr($line, 9));
                                $repl = md5(random_bytes(32));
                                PandaRoseUI::$repl[$repl] = pandarose($incl);
                                $line = $repl;
                            }
                            $content .= $line . "\n";
                        }
                        $parsedown = new Parsedown();
                        $output = $parsedown->text($content);
                        break;
                    case 'json':
                        if (!$json = json_decode($content, true)) {
                            $json = [];
                        }
                    case 'yml':
                    case 'yaml':
                        $result = ($json !== null)
                            ? $json : Yaml::parse($content);
                        if (isset($result['render'])) {
                            $render = $result['render'];
                            unset($result['render']);
                            PandaRoseUI::$test = 1;
                            $output = pandarose($render, $result);
                        }
                        break;
                    case 'scss':
                        $scss = new Compiler();
                        $scss->setVariables([]);
                        $content = $scss->compile($content);
                    case 'css':
                        $output = '<style>' . $content . '</style>';
                        break;
                    default:
                        $output = parseTwig($content, $data);
                }
            } else {
                $output = parseTwig($content, $data);
            }
        }
        if ($json) {
            ob_clean();
        }
        $verb = $_SERVER['REQUEST_METHOD'];
        $responders = [];
        $v = ['state' => []];
        $raw = false;
        $script = '';
        if (is_iterable($result)) {
            foreach ($result as $q => $r) {
                if ($q === 'raw' and $r === true)
                    $raw = true;
                else if ($q === 'script')
                    $script = $r;
                else if ($q === 'final')
                    PandaRoseUI::$final[$filepath] = $r;
                else if (is_string($q) and substr($q, 0, 5) === 'state')
                    if ($verb === 'GET')
                        PandaRoseUI::keyString($q, $r, $v);
                $responders[$q] = $r;
            }
            if ($result instanceof Generator) {
                if ($ret = $result->getReturn())
                    $responders[$q] = $r;
            }
        }

        if ($raw and PandaRoseUI::$deep > 1) return '';

        if (is_callable($result))
            $responders[] = $result;
        foreach ($responders as $key => $result) {
            $keyVerb = 'POST';
            if (!is_string($key)) $key = '';
            if (substr_count($key, '?')) {
                list($keyVerb, $key) = explode('?', $key);
            }
            if (strtoupper($key) === $key) {
                $keyVerb = $key;
                $key = '';
            }
            parse_str($key, $bits);
            if (count($bits)) {
                foreach ($bits as $bit => $val) {
                    if ($val === '' && isset($in[$bit])) {
                    } else if ($val && isset($in[$bit]) and $in[$bit] == $val) {
                    } else {
                        continue 2;
                    }
                }
            }
            $reflect = new ReflectionFunction($result);
            $args = [];
            foreach ($reflect->getParameters() as $param) {
                $name = $param->getName();
                if ($name === 'in') {
                    $args[] = $in;
                } else {
                    if (isset($in[$name]))
                        $args[] = $in[$name];
                    else if ($param->allowsNull())
                        $args[] = null;
                }
            }
            try {
                $result = $result(...$args);
                if ($verb !== 'GET') {
                    if (is_iterable($result)) {
                        again:
                        foreach ($result as $q => $r) {
                            if (is_string($q) and substr($q, 0, 5) === 'state') {
                                PandaRoseUI::keyString($q, $r, $v);
                            } else if (is_string($q)) {
                                $q = 'state.' . $q;
                                PandaRoseUI::keyString($q, $r, $v);
                            }
                        }
                        if ($result instanceof Generator) {
                            if ($ret = $result->getReturn()) {
                                $result = $ret;
                                if (is_iterable($ret))
                                    goto again;
                            }
                        }
                    }
                    $page = [
                        'data' => count($v['state']) ? $v['state'] : new stdClass
                    ];
                    if (is_string($result))
                        $page['__html'] = base64_encode($result);
                    header('Content-Type: application/json');
                    echo json_encode([
                        'page' => $page
                    ]);
                    exit;
                }
            } catch (\Throwable $e) {
            }
        }

        $html = $newHtml ?? $output;

        if ($json) {
            header('Content-Type: application/json');
            echo json_encode([
                'page' => [
                    'html' => $html
                ]
            ]);
            exit;
        }

        if ($verb === 'GET') {
            if (
                !$raw and
                !substr_count($html, '<pandarose-ui') and
                PandaRoseUI::$deep < 2
            ) {
                $html = <<<HTML
                <pandarose-ui style="display:none">{$html}</pandarose-ui>
                HTML;
            }

            if ($raw and $script) {
                $html .= '<script>' .
                    'var jq = jQuery;' .
                    'jq(document).ready(function() {' . $script . '});' .
                    '</script>';
            }

            PandaRoseUI::$deep--;

            return $html;
        }
    }

    PandaRoseUI::$deep--;
}

function pandarose_ui_shortcode($args, $content = '')
{
    wp_enqueue_script("jquery");
    PandaRoseUI::enqueue();

    if (!is_array($args)) $args = [];

    return do_shortcode(pandarose($args, null, $content));
}

add_action('admin_menu', 'pandarose_ui_admin');

function pandarose_ui_admin()
{
    add_menu_page('PandaRose UI', 'PandaRose UI', 'manage_options', 'pandarose-ui', 'admin_page');
}

function admin_page()
{
    PandaRoseUI::enqueue();
    $admin = isset($_GET['admin']) ? $_GET['admin'] : 'default';
    if ($admin = (preg_replace('/[^_a-z\d]/i', '', $admin))) {
        $filepath = __DIR__ . DIRECTORY_SEPARATOR . 'admin' . DIRECTORY_SEPARATOR . $admin . '.php';
        echo <<<JS
        <script>
            window.PushHandler = function(path) {
                path = window.FoundationHelpers.lodash.trim(path, '/');
                window.FoundationHelpers.insertUrlParam('admin', path, true);
            }
        </script>
        JS;
        if ($result = pandarose_template($filepath)) {
            echo $result;
        } else {
            echo 'not found';
        }
    } else {
        echo 'Not Found';
    }
}

function pandarose_post_type($suffix = '', $opts = [])
{
    if (is_string($opts))
        $opts = ['name' => $opts];

    if ($suffix and (!isset($opts['top']) or !$opts['top'])) {
        $opts['show_in_menu'] = 'edit.php?post_type=pandarose';
        if (!isset($opts['name']))
            $opts['name'] = ucfirst(strtolower($suffix));
        if (isset($opts['top']))
            unset($opts['top']);
    }

    if (!isset($opts['name']))
        $opts['name'] = 'UI Templates';

    $labels = [
        'name'          => $opts['name'],
        'singular_name' => isset($opts['singular_name'])
            ? $opts['singular_name'] : $opts['name'],
        'menu_name'     => isset($opts['menu_name'])
            ? $opts['menu_name'] : $opts['name'],
    ];

    if (isset($opts['singular_name'])) unset($opts['singular_name']);
    if (isset($opts['menu_name'])) unset($opts['menu_name']);
    unset($opts['name']);

    $opts = array_merge([
        'labels'             => $labels,
        'public'             => false,
        'publicly_queryable' => false,
        'show_ui'            => true,
        'show_in_menu'       => true,
        'query_var'          => true,
        'capability_type'    => 'post',
        'has_archive'        => true,
        'hierarchical'       => false,
        'menu_position'      => null,
        'supports'           => ['title', 'editor', 'custom-fields', 'revisions', 'comments', 'thumbnail'],
    ], $opts);

    register_post_type($suffix
        ? 'pandarose-' . $suffix
        : 'pandarose', $opts);
}

function pandarose_ui_input_shortcode($args)
{
    $parts = ['input', 'type' => 'text'];

    foreach ($args as $k => $v) {
        $parts[$k] = $v;
    }

    $out = '';

    foreach ($parts as $k => $v) {
        if (is_numeric($k)) {
            $out .= $v . ' ';
        } else {
            $out .= $k . '="' . addslashes($v) . '" ';
        }
    }

    return '<' . trim($out) . '>';
}

add_action('admin_enqueue_scripts', [PandaRoseUI::class, 'enqueue']);
add_shortcode('pandarose_ui', 'pandarose_ui_shortcode');
add_shortcode('pandarose-ui', 'pandarose_ui_shortcode');
add_shortcode('pandarose', 'pandarose_ui_shortcode');
add_shortcode('ui', 'pandarose_ui_shortcode');
add_shortcode('ui_input', 'pandarose_ui_input_shortcode');
add_shortcode('ui-input', 'pandarose_ui_input_shortcode');

add_action('init', function () {
    pandarose_post_type();
    pandarose_post_type('form');
});

function pandarose_inject_editor($content, $opts = [])
{
    $opts = array_merge([
        'textLabel' => 'Code'
    ], $opts);

    PandaRoseUI::after(
        function ($out) use ($content, $opts) {
            if (!strpos($out, '<div id="wp-content-editor-container"'))
                return $out;
            $out = <<<HTML
                <script>
                    jQuery('#wp-content-editor-container').hide();
                    jQuery('#content-tmce').hide();
                </script>
                <style>
                    #wp-content-editor-container { 
                        display: none !important;
                    }
                    #content-tmce {
                        display: none !important;
                    }
                    #content-html {
                        opacity: 0;
                    }
                    #wp-word-count {
                        display: none !important;
                    }
                </style>
                HTML . $out;
            $pos = strpos($out, '<div id="wp-content-editor-container"');
            $out = substr($out, 0, $pos)
                . '<kojo-foundation>' . $content . '</kojo-foundation>'
                . substr($out, $pos, strlen($out) - strlen($pos));
            $out .= <<<HTML
            <script>
                jQuery('#content-html').text("{$opts['textLabel']}").css('opacity', 1);
            </script>
            HTML;
            return $out;
        }
    );
}

function pandarose_monaco($language = 'twig', $opts = [])
{
    pandarose_inject_editor(
        <<<HTML
        <div id="pandarose-editor" style="opacity: 0; min-height: 500px">
            <textarea use="monaco" language="$language" bind="code" expr="typeof bound.value !== 'undefined'" width="100%" height="500">
            </textarea>
        </div>
        <script code-bind="code">
            switch(changed) {
                case 'code':
                    if(self.started)
                        jQuery('textarea.wp-editor-area').val(bound.code);
                    self.started = true;
                    break;
                default:
                    var code = jQuery('textarea.wp-editor-area').val();
                    mutate.code.change(code || '');
                    jQuery('#pandarose-editor').css('opacity', 1);
            }
        </script>
        HTML,
        $opts
    );
}

add_filter('wp_default_editor', function ($editor) {
    $type = get_post_type();
    if ($type === 'pandarose' or substr($type, 0, 10) === 'pandarose-') {
        $language = 'twig';
        if ($post = get_post()) {
            $title = $post->post_title;
            $extn = pathinfo($title, PATHINFO_EXTENSION);
            $opts = [];
            switch ($extn) {
                case 'yml':
                case 'yaml':
                    $language = 'yaml';
                    break;
                case 'js':
                case 'jsx':
                    $language = 'javascript';
                    break;
                case 'md':
                    $language = 'markdown';
                    break;
                case 'css':
                    $language = 'css';
                    break;
                case 'scss':
                    $language = 'scss';
                    break;
                case 'json':
                    $language = 'json';
                    break;
            }
        }
        pandarose_monaco($language, $opts);
        return 'html';
    }
    return 'tinymce';
});


add_action(
    'shutdown',
    [PandaRoseUI::class, 'sendBuffer'],
    0
);

ob_start();
